@@ -3,9 +3,9 @@
 Module      : HsLua.Core.Primary
 Copyright   : © 2007–2012 Gracjan Polak;
               © 2012–2016 Ömer Sinan Ağacan;
-              © 2017-2022 Albert Krewinkel
+              © 2017-2024 Albert Krewinkel
 License     : MIT
-Maintainer  : Albert Krewinkel <tarleb+hslua@zeitkraut.de>
+Maintainer  : Albert Krewinkel <tarleb@hslua.org>
 Stability   : beta
 Portability : non-portable (depends on GHC)
 
@@ -98,7 +98,7 @@ absindex = liftLua1 lua_absindex
 -- to its original configuration. This is considered good programming
 -- practice.
 --
--- See <https://www.lua.org/manual/5.3/manual.html#lua_call lua_call>.
+-- See <https://www.lua.org/manual/5.4/manual.html#lua_call lua_call>.
 call :: LuaError e => NumArgs -> NumResults -> LuaE e ()
 call nargs nresults = do
   res <- pcall nargs nresults Nothing
@@ -145,7 +145,7 @@ close = lua_close
 --    LE: compares for less or equal (<=)
 --
 -- Wraps 'hslua_compare'. See also
--- <https://www.lua.org/manual/5.3/manual.html#lua_compare lua_compare>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_compare lua_compare>.
 compare :: LuaError e
         => StackIndex         -- ^ idx1
         -> StackIndex         -- ^ idx2
@@ -160,11 +160,11 @@ compare idx1 idx2 relOp = fromLuaBool <$!> liftLuaThrow
 -- value on the stack (that is, the function does nothing); if @n@ is 0,
 -- the result is the empty string. Concatenation is performed following
 -- the usual semantics of Lua (see
--- <https://www.lua.org/manual/5.3/manual.html#3.4.6 §3.4.6> of the Lua
+-- <https://www.lua.org/manual/5.4/manual.html#3.4.6 §3.4.6> of the Lua
 -- manual).
 --
 -- Wraps 'hslua_concat'. See also
--- <https://www.lua.org/manual/5.3/manual.html#lua_concat lua_concat>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_concat lua_concat>.
 concat :: LuaError e => NumArgs -> LuaE e ()
 concat n = liftLuaThrow (`hslua_concat` n)
 {-# INLINABLE concat #-}
@@ -219,20 +219,21 @@ error = liftLua hslua_error
 --
 -- Wraps 'lua_gc'.
 gc :: GCControl -> LuaE e Int
-gc what = liftLua $ \l ->
-  fromIntegral <$!> lua_gc l (toGCcode what) (toGCdata what)
+gc what = liftLua $ \l -> do
+  let (data1, data2, data3) = toGCdata what
+  fromIntegral <$!> lua_gc l (toGCcode what) data1 data2 data3
 {-# INLINABLE gc #-}
 
 -- | Pushes onto the stack the value @t[k]@, where @t@ is the value at
 -- the given stack index. As in Lua, this function may trigger a
 -- metamethod for the "index" event (see
--- <https://www.lua.org/manual/5.3/manual.html#2.4 §2.4> of Lua's
+-- <https://www.lua.org/manual/5.4/manual.html#2.4 §2.4> of Lua's
 -- manual).
 --
 -- Errors on the Lua side are propagated.
 --
 -- See also
--- <https://www.lua.org/manual/5.3/manual.html#lua_getfield lua_getfield>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_getfield lua_getfield>.
 getfield :: LuaError e => StackIndex -> Name -> LuaE e Type
 getfield i (Name s) = do
   absidx <- absindex i
@@ -267,13 +268,13 @@ getmetatable n = liftLua $ \l ->
 -- This function pops the key from the stack, pushing the resulting
 -- value in its place. As in Lua, this function may trigger a metamethod
 -- for the "index" event (see
--- <https://www.lua.org/manual/5.3/manual.html#2.4 §2.4> of Lua's
+-- <https://www.lua.org/manual/5.4/manual.html#2.4 §2.4> of Lua's
 -- manual).
 --
 -- Errors on the Lua side are caught and rethrown.
 --
 -- Wraps 'hslua_gettable'. See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_gettable lua_gettable>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_gettable lua_gettable>.
 gettable :: LuaError e => StackIndex -> LuaE e Type
 gettable n = liftLuaThrow (\l -> fmap toType . hslua_gettable l n)
 {-# INLINABLE gettable #-}
@@ -287,14 +288,19 @@ gettop :: LuaE e StackIndex
 gettop = liftLua lua_gettop
 {-# INLINABLE gettop #-}
 
--- | Pushes onto the stack the Lua value associated with the full
--- userdata at the given index.
+-- | Pushes onto the stack the @n@-th user value associated with the
+-- full userdata at the given index and returns the type of the pushed
+-- value.
 --
--- Returns the type of the pushed value.
+-- If the userdata does not have that value, pushes __nil__ and returns
+-- 'LUA_TNONE'.
 --
--- Wraps 'lua_getuservalue'.
-getuservalue :: StackIndex -> LuaE e Type
-getuservalue idx = toType <$> liftLua (`lua_getuservalue` idx)
+-- Wraps 'lua_getiuservalue'.
+getiuservalue :: StackIndex  -- ^ index
+             -> Int         -- ^ n
+             -> LuaE e Type
+getiuservalue idx n = liftLua $ \l ->
+  toType <$!> lua_getiuservalue l idx (fromIntegral n)
 
 -- | Moves the top element into the given valid index, shifting up the
 -- elements above this index to open space. This function cannot be
@@ -442,7 +448,7 @@ lessthan index1 index2 = compare index1 index2 LT
 --
 -- The @chunkname@ argument gives a name to the chunk, which is used for
 -- error messages and in debug information (see
--- <https://www.lua.org/manual/5.3/manual.html#4.9 §4.9>). Note that the
+-- <https://www.lua.org/manual/5.4/manual.html#4.7 §4.7>). Note that the
 -- @chunkname@ is used as a C string, so it may not contain null-bytes.
 --
 -- This is a wrapper of 'lua_load'.
@@ -464,19 +470,29 @@ ltype idx = toType <$!> liftLua (`lua_type` idx)
 -- equivalent to @createtable 0 0@.
 --
 -- See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_newtable lua_newtable>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_newtable lua_newtable>.
 newtable :: LuaE e ()
 newtable = createtable 0 0
 {-# INLINABLE newtable #-}
 
--- | This function allocates a new block of memory with the given size,
--- pushes onto the stack a new full userdata with the block address, and
--- returns this address. The host program can freely use this memory.
+-- | This function creates and pushes on the stack a new full userdata,
+-- with @nuvalue@ associated Lua values, called @user values@, plus an
+-- associated block of raw memory with @size@ bytes. (The user values
+-- can be set and read with the functions 'lua_setiuservalue' and
+-- 'lua_getiuservalue'.)
 --
--- This function wraps 'lua_newuserdata'.
-newuserdata :: Int -> LuaE e (Ptr ())
-newuserdata = liftLua1 lua_newuserdata . fromIntegral
-{-# INLINABLE newuserdata #-}
+-- The function returns the address of the block of memory. Lua ensures
+-- that this address is valid as long as the corresponding userdata is
+-- alive (see <https://www.lua.org/manual/5.4/manual.html#2.5 §2.5>).
+-- Moreover, if the userdata is marked for finalization (see
+-- <https://www.lua.org/manual/5.4/manual.html#2.5.3 §2.5.3>), its
+-- address is valid at least until the call to its finalizer.
+--
+-- This function wraps 'lua_newuserdatauv'.
+newuserdatauv :: Int {- ^ size -} -> Int {- ^ nuvalue -} -> LuaE e (Ptr ())
+newuserdatauv size nuvalue =
+  liftLua $ \l -> lua_newuserdatauv l (fromIntegral size) (fromIntegral nuvalue)
+{-# INLINABLE newuserdatauv #-}
 
 -- | Pops a key from the stack, and pushes a key–value pair from the
 -- table at the given index (the "next" pair after the given key). If
@@ -487,7 +503,7 @@ newuserdata = liftLua1 lua_newuserdata . fromIntegral
 --
 -- This function wraps 'hslua_next'.
 -- See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_next lua_next>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_next lua_next>.
 next :: LuaError e => StackIndex -> LuaE e Bool
 next idx = fromLuaBool <$!> liftLuaThrow (\l -> hslua_next l idx)
 {-# INLINABLE next #-}
@@ -585,7 +601,7 @@ pcall nargs nresults msgh = liftLua $ \l ->
 
 -- | Pops @n@ elements from the stack.
 --
--- See also: <https://www.lua.org/manual/5.3/manual.html#lua_pop lua_pop>.
+-- See also: <https://www.lua.org/manual/5.4/manual.html#lua_pop lua_pop>.
 pop :: Int -> LuaE e ()
 pop n = liftLua $ \l -> lua_pop l (fromIntegral n)
 {-# INLINABLE pop #-}
@@ -625,7 +641,7 @@ pushcclosure f n = liftLua $ \l -> lua_pushcclosure l f n
 -- to receive its parameters and return its results (see @'CFunction'@)
 --
 -- Same as @flip 'pushcclosure' 0@.
--- <https://www.lua.org/manual/5.3/manual.html#lua_pushcfunction lua_pushcfunction>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_pushcfunction lua_pushcfunction>.
 pushcfunction :: CFunction -> LuaE e ()
 pushcfunction f = pushcclosure f 0
 {-# INLINABLE pushcfunction #-}
@@ -790,7 +806,7 @@ replace n = liftLua $ \l ->  lua_replace l n
 -- This function cannot be called with a pseudo-index, because a
 -- pseudo-index is not an actual stack position.
 --
--- <https://www.lua.org/manual/5.3/manual.html#lua_rotate>
+-- <https://www.lua.org/manual/5.4/manual.html#lua_rotate>
 rotate :: StackIndex  -- ^ @idx@
        -> Int         -- ^ @n@
        -> LuaE e ()
@@ -802,13 +818,13 @@ rotate idx n = liftLua $ \l -> lua_rotate l idx (fromIntegral n)
 --
 -- This function pops the value from the stack. As in Lua, this function
 -- may trigger a metamethod for the "newindex" event (see
--- <https://www.lua.org/manual/5.3/manual.html#2.4 §2.4> of the Lua 5.3
+-- <https://www.lua.org/manual/5.4/manual.html#2.4 §2.4> of the Lua 5.4
 -- Reference Manual).
 --
 -- Errors on the Lua side are caught and rethrown as a @'Exception'@.
 --
 -- See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_setfield lua_setfield>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_setfield lua_setfield>.
 setfield :: LuaError e => StackIndex -> Name -> LuaE e ()
 setfield i (Name s) = do
   absidx <- absindex i
@@ -823,7 +839,7 @@ setfield i (Name s) = do
 -- Errors on the Lua side are caught and rethrown as 'Exception'.
 --
 -- Wraps 'hslua_setglobal'. See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_setglobal lua_setglobal>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_setglobal lua_setglobal>.
 setglobal :: LuaError e => Name {- ^ name -} -> LuaE e ()
 setglobal (Name name) = liftLuaThrow $ \l status' ->
   B.unsafeUseAsCStringLen name $ \(namePtr, nameLen) ->
@@ -844,8 +860,8 @@ setmetatable idx = liftLua $ \l -> lua_setmetatable l idx
 --
 -- This function pops both the key and the value from the stack. As in
 -- Lua, this function may trigger a metamethod for the "newindex" event
--- (see <https://www.lua.org/manual/5.3/manual.html#2.4 §2.4> of the Lua
--- 5.3 Reference Manual).
+-- (see <https://www.lua.org/manual/5.4/manual.html#2.4 §2.4> of the Lua
+-- 5.4 Reference Manual).
 --
 -- Errors on the Lua side are caught and rethrown.
 --
@@ -863,12 +879,23 @@ settop :: StackIndex -> LuaE e ()
 settop = liftLua1 lua_settop
 {-# INLINABLE settop #-}
 
--- | Pops a value from the stack and sets it as the new value associated
--- to the full userdata at the given index.
+-- | Pops a value from the stack and sets it as the new @n@-th user
+-- value associated to the full userdata at the given index. Returns 0
+-- if the userdata does not have that value.
 --
--- <https://www.lua.org/manual/5.3/manual.html#lua_setuservalue>
-setuservalue :: StackIndex -> LuaE e ()
-setuservalue idx = liftLua (`lua_setuservalue` idx)
+-- Wraps 'lua_setiuservalue'.
+setiuservalue :: StackIndex {- ^ index -} -> Int {- ^ n -} -> LuaE e Bool
+setiuservalue idx n = liftLua $ \l ->
+  fromLuaBool <$!> lua_setiuservalue l idx (fromIntegral n)
+
+-- | Sets the warning function to be used by Lua to emit warnings (see
+-- 'WarnFunction'). The @ud@ parameter sets the value @ud@ passed to the
+-- warning function.
+setwarnf :: WarnFunction -- ^ f
+         -> Ptr ()       -- ^ ud
+         -> LuaE e ()
+setwarnf f ud = liftLua $ \l ->
+  lua_setwarnf l f ud
 
 -- |  Returns the status of this Lua thread.
 --
@@ -909,14 +936,14 @@ tocfunction n = liftLua $ \l -> do
 -- | Converts the Lua value at the given acceptable index to the signed
 -- integral type 'Lua.Integer'. The Lua value must be an integer, a
 -- number or a string convertible to an integer (see
--- <https://www.lua.org/manual/5.3/manual.html#3.4.3 §3.4.3> of the Lua
--- 5.3 Reference Manual); otherwise, @tointeger@ returns @Nothing@.
+-- <https://www.lua.org/manual/5.4/manual.html#3.4.3 §3.4.3> of the Lua
+-- 5.4 Reference Manual); otherwise, @tointeger@ returns @Nothing@.
 --
 -- If the number is not an integer, it is truncated in some
 -- non-specified way.
 --
 -- Wraps 'lua_tointegerx'. See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_tointeger lua_tointeger>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_tointeger lua_tointeger>.
 tointeger :: StackIndex -> LuaE e (Maybe Lua.Integer)
 tointeger n = liftLua $ \l -> alloca $ \boolPtr -> do
   res <- lua_tointegerx l n boolPtr
@@ -929,7 +956,7 @@ tointeger n = liftLua $ \l -> alloca $ \boolPtr -> do
 -- otherwise, @tonumber@ returns @'Nothing'@.
 --
 -- Wraps 'lua_tonumberx'. See also
--- <https://www.lua.org/manual/5.3/manual.html#lua_tonumber lua_tonumber>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_tonumber lua_tonumber>.
 tonumber :: StackIndex -> LuaE e (Maybe Lua.Number)
 tonumber n = liftLua $ \l -> alloca $ \bptr -> do
   res <- lua_tonumberx l n bptr
@@ -1006,11 +1033,11 @@ typename tp = liftLua $ \l ->
 {-# INLINABLE typename #-}
 
 -- | Returns the pseudo-index that represents the @i@-th upvalue of the
--- running function (see <https://www.lua.org/manual/5.3/manual.html#4.4
--- §4.4> of the Lua 5.3 reference manual).
+-- running function (see <https://www.lua.org/manual/5.4/manual.html#4.2
+-- §4.2> of the Lua 5.4 reference manual).
 --
 -- See also:
--- <https://www.lua.org/manual/5.3/manual.html#lua_upvalueindex lua_upvalueindex>.
+-- <https://www.lua.org/manual/5.4/manual.html#lua_upvalueindex lua_upvalueindex>.
 upvalueindex :: StackIndex -> StackIndex
 upvalueindex i = registryindex - i
 {-# INLINABLE upvalueindex #-}
